#ifndef __LSM_SOLVER_H
#define __LSM_SOLVER_H
#include <dolfin/Solver.h>
#include "forms/LSM_adv_CN.h"
#include "forms/LSM_grad.h"
#include "forms/LSM_re.h"

#include "Mesh_fcns.hpp"
#include "Diff_fcns.hpp"
#include "Vec_Func_fcns.hpp"

namespace dolfin
{

  class LSM_Solver : public Solver
  {
  public:

    LSM_Solver(Mesh&, Function&, real, real);

    void solve();
    Vector& timestep(Vector&);
    Vector& timestep(real, Vector&, Function&);
    Vector& timestep_only_reinit(Vector&,real);

  private:

    Mesh& mesh;
    Function& vf;



    real dt;
    real min_cell;

    int iter_re;
    int iter_re_max;
    real tau;
    real dtau;
    real x_re_norm;
    real err_rel;

    real epsilon;

    real diff_coeff;

    real normal_cutoff;

    KrylovSolver k_solver;
    LU lu_solver;

    Matrix A_adv;
    Vector x, x0, b_adv;

    Matrix A_re;
    Vector b_re;

    Matrix A_grad;
    Vector b_grad, norml_vec;
    Function* phi_norml;

    Matrix A_smooth;
    Vector b_smooth;
    Function* phi_smooth; 

    unsigned int Nv;

    LSM_adv_CN::BilinearForm::TrialElement element_phi;
    LSM_grad::BilinearForm::TrialElement element_grad;

    Function* phi; 
    Function* phi_0;

    BilinearForm* a_adv;
    LinearForm* L_adv;

    BilinearForm* a_re;
    LinearForm* L_re;

    Vec_Func_fcns vff;
    Diff_fcns df;
    Mesh_fcns mf;
  };
}
#endif
