// Vector and Function functions for
// C-LSM: Conservative Level Set Method
//
// Created: 2006-09-13
// Dag Lindbo, dag@f.kth.se

#include "Vec_Func_fcns.hpp"

using namespace dolfin;

Vec_Func_fcns::Vec_Func_fcns()
{}

// THIS IS NOT YET WORKING (?)
// COMPUTE L2-NORM OF DIFF UF AND DF ----------------------------%%%
real Vec_Func_fcns::L2_Error_Exact(Mesh& mesh, 
				   Function& exact, 
				   Function& computed)
{
  Vector p = computed.vector();
  Function comp(p,mesh,computed.element());

  Scalar_L2Error::LinearForm L(exact, computed); // causes sigsegv later
  //Scalar_L2Error::LinearForm L(exact, comp); // works
  Vector qq;
  FEM::assemble(L, qq, mesh);
  real norm = sqrt(qq.sum());
  dolfin_info("L2 error: %.3e", norm);

  return norm;
}

// THIS IS NOT YET WORKING (?)
// COMPUTE L2-NORM OF DIFF DF AND DF ----------------------------%%%
real Vec_Func_fcns::L2_Error_DF_DF(Mesh& mesh, 
				   Function& f1, 
				   Function& f2)
{
  Vector fun1_vec = f1.vector();
  Vector fun2_vec = f2.vector();
 
  Function fun1(fun1_vec,mesh,f1.element());
  Function fun2(fun2_vec,mesh,f2.element());

  Scalar_L2Error::LinearForm L(fun1, fun2);
  Vector qq;
  FEM::assemble(L, qq, mesh);
  real norm = sqrt(qq.sum());
  dolfin_info("L2 error: %.3e", norm);

  return norm;
}

// INTEGRATES A SCALAR FUNCTION OVER ITS MESH -------------------%%%
real Vec_Func_fcns::integrate_fcn(Mesh& mesh, Function& f)
{
  Vector f_vec = f.vector();
  Function func(f_vec,mesh,f.element());

  Scalar_Integrate::LinearForm L(func);
  Vector qq;
  FEM::assemble(L, qq, mesh);

  real inte = qq.sum();
  dolfin_info("Integral: %f",inte);

  return inte;
}

// THIS IS NOT YET ACCURATE (?)
// COMPUTES THE AREA WITHIN 0.5-CONTOUR OF PHI ----------------&&&
real Vec_Func_fcns::integrateLevelSet(Mesh& mesh, 
			       Function& phi)
{
  Vector q;
  Vector p = phi.vector();
  int len = p.size();

  for(int i = 0; i<len;i++)
    p(i) = (p(i)>0.5) ? 1:0; 

  Function f(p,mesh,phi.element());

  Scalar_Integrate::LinearForm L(f);
  FEM::assemble(L,q,mesh);

  real area = q.sum();
  dolfin_info("Integral over level set: %f",area);

  return area;
}

// THIS IS NOT YET ACCURATE (!)
// COMPUTES THE AREA WITHIN 0.5-CONTOUR OF PHI ----------------&&&
real Vec_Func_fcns::integrateLevelSet2(Mesh& mesh, 
				Function& phi)
{
  real tol = 1e-1;
  real a;

  Vector q;

  Vector p = phi.vector();
  Vector pp = phi.vector();

  int len = p.size();

  for(int i = 0; i<len;i++)
    {
      a = p(i)-0.5;
      if(fabs(a) < tol)
	{
	  p(i) = 1.0/tol;
	  
	  if(a < 0)
	    pp(i) = -tol;
	  else
	    pp(i) = tol;
	}
      else
	{
	  p(i) = 1.0/fabs(a);
	  pp(i) = a;
	}
    }

  // f = phi-0.5
  Function f(pp,mesh,phi.element());

  // g = 1/|phi-0.5|
  Function g(p,mesh,phi.element());

  Scalar_Integrate_prod::LinearForm L(f,g);
  FEM::assemble(L,q,mesh);

  real area = q.sum();
  dolfin_info("Integral over level set: %f",area);

  return area;
}

// COMPUTES RELATIVE ERROR BETWEEN TWO VECTORS ------------------%%%
real Vec_Func_fcns::computeRelError(Vector& x, Vector& x0)
{
  real x_norm = 0;
  real err_rel = 0;

  for(unsigned int i=0;i<x.size();i++)
    {
      x_norm += fabs(x(i));
      err_rel += fabs(x(i)-x0(i));
    }

  return err_rel/x_norm;
}

// RESTRICTS A USER FUNCTION ONTO A DISCRETE FUNCTION -----------%%%
Vector& Vec_Func_fcns::restrictUFtoDF(Mesh& mesh , Function& p)
{
  Matrix A;
  Vector b;

  LSM_restrict::BilinearForm a_res;
  LSM_restrict::LinearForm L_res(p);
  FEM::assemble(a_res, L_res, A, b, mesh);
  lu_solver.solve(A,res_vec,b);

  return res_vec;
}
// MAXIMUM VALUE IN A VECTOR ------------------------------------%%%
real Vec_Func_fcns::maxVal(Vector& x)
{
  real maxval = -1e10;
  int sz = x.size();

  for(int i = 0; i<sz; i++)
    {
      if(x(i) > maxval) maxval = x(i);
    }

  return maxval;
}

// MAXIMUM MAGNITUDE IN A VECTOR VALUED FUNC -------------------%%%
real Vec_Func_fcns::maxMag(Vector& x)
{
  real maxmag = -1e10;
  real mag;
  int sz = x.size()/2;

  for(int i = 0; i<sz; i++)
    {
      mag = sqrt(x(i)*x(i) + x(i+sz)*x(i+sz));
      if(mag > maxmag) maxmag = mag;
    }

  return maxmag;
}
